import uvicore
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.foundation.events import app as AppEvents


@uvicore.provider()
class Cache(ServiceProvider):

    def register(self) -> None:
        # Register event listeners
        #AppEvents.Registered.listen(bootstrap.Cache)

        #self.bind_override('uvicore.cache.cache.Cache', 'uvicore.cache.cache2.Cache2')
        # self.bind('uvicore.cache.cache.Cache', 'uvicore.cache.cache.Cache',
        #     aliases=['cache0', 'cache'],
        #     singleton=False,
        # )

        #self.bind('uvicore.cache.cache.Cache')

        # self.bind('uvicore.cache.cache.Cache', 'uvicore.cache.cache2.Cache2',
        #     aliases=['cacheO'],
        #     singleton=True,
        # )

        # Set uvicore.log global connecting to default store
        uvicore.cache = uvicore.ioc.make('uvicore.cache.manager.Manager').connect()


    def boot(self) -> None:
        # Define service provider registration control
        #self.registers(self.package.config.registers)

        # Import cache to fire up Ioc so we can later use as short 'cache' names
        #from uvicore.cache.cache import Cache
        pass



import uvicore
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.console.provider import Cli
from uvicore.http.provider import Http
from uvicore.foundation.events import app as AppEvents
from uvicore.http import bootstrap


@uvicore.provider()
class Http(ServiceProvider, Cli, Http):

    def register(self) -> None:
        """Register package into uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs and IoC bindings.  Configs are deep merged only after
        all packages are registered.  No real work should be performed here as it
        is very early in the bootstraping process and most internal processes are not
        instantiated yet.
        """

        # # Register events used in this package
        # self.events.register(
        #     name='uvicore.http.events.server.Startup',
        #     description='HTTP Server has been started.  This is the Starlette startup event.',
        #     is_async=False,
        # )
        # self.events.register(
        #     name='uvicore.http.events.server.Shutdown',
        #     description='HTTP Server has been shutdown.  This is the Starlette shutdown event.',
        #     is_async=False,
        # )

        # Register events used in this package
        # self.events.register({
        #     # This is a Starlette to Uvicore event
        #     'uvicore.http.events.server.Startup': {
        #         'description': 'HTTP Server has been started.  This is the Starlette startup event.',
        #         'type': 'string',
        #     },
        #     # This is a Starlette to Uvicore event
        #     'uvicore.http.events.server.Shutdown': {
        #         'description': 'HTTP Server has been shutdown.  This is the Starlette shutdown event.',
        #         'type': 'string',
        #     },

        # })


        #     @uvicore.app.http.on_event("shutdown")
        #     async def shutdown():
        #         for database in self.databases.values():
        #             await database.disconnect()

        # Register event listeners
        # After all providers are booted we have a complete list of view paths
        # and template options fully merged.  Now we can fire up the static
        # paths and template system.
        # We handle this event even in CONSOLE mode to handle edge cases.
        AppEvents.Booted.listen(bootstrap.Http)


        # Register IoC bindings only if running in HTTP mode
        #if self.app.is_http:

            # # Bind HTTP Server
            # self.bind('Http', 'uvicore.http.server._Server',
            #     aliases=['uvicore.http.server.Server', 'http', 'HTTP'],
            #     singleton=True,
            #     kwargs={
            #         'debug': uvicore.config('app.debug'),
            #         'title': uvicore.config('app.openapi.title'),
            #         'version': uvicore.app.version,
            #         'openapi_url': uvicore.config('app.openapi.url'),
            #         'docs_url': uvicore.config('app.openapi.docs_url'),
            #         'redoc_url': uvicore.config('app.openapi.redoc_url'),
            #     }
            # )
            # No because I added default to make
            #self.bind('WebRouter', 'uvicore.http.routing.web_router._WebRouter', aliases=['uvicore.http.routing.web_router.WebRouter', 'web_router'])
            #self.bind('ApiRouter', 'uvicore.http.routing.api_router._ApiRouter', aliases=['uvicore.http.routing.api_router.ApiRouter', 'api_router'])
            #self.bind('Routes', 'uvicore.http.routing.routes._Routes', aliases=['uvicore.http.routing.routes.Routes', 'routes'])
            #self.bind('StaticFiles', 'uvicore.http.static._StaticFiles', aliases=['Static', 'static'])

            # Default templating system is Jinja2.  Users can overwrite this
            # easily in their app configs 'bindings' dictionary.
            #self.bind('Templates', 'uvicore.http.templating.jinja.Jinja', singleton=True, aliases=['templates'])

            # Fire up a FastAPI server instance and store on app.http
            #self.app._http = uvicore.ioc.make('uvicore.http.server._Server')
            #self.app._http = uvicore.ioc.make('uvicore.http.servers.api._Server')  # FastAPI as Base
            #self.app._http = uvicore.ioc.make('uvicore.http.servers.web._Server')  # Starlette as Base

            # Base HTTP Server
            #self.app._http = uvicore.ioc.make('uvicore.http.servers.api._Server')  # FastAPI as Base
            # self.app._http: _FastAPI = _FastAPI(
            #     debug=uvicore.config('app.debug'),
            #     title=uvicore.config('app.openapi.title'),
            #     version=uvicore.app.version,
            #     openapi_url=uvicore.config('app.openapi.url'),
            #     docs_url=uvicore.config('app.openapi.docs_url'),
            #     redoc_url=uvicore.config('app.openapi.redoc_url'),
            #     root_path='/api',
            # )

    def boot(self) -> None:
        """Bootstrap package into uvicore framework.
        Boot takes place after all packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all configs.
        This is where you load views, assets, routes, commands...
        """
        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define HTTP Middleware
        #self.middleware(uvicore.config('app.middleware'))

        # Define CLI commands
        self.commands({
            'http': {
                'help': 'Uvicore HTTP Commands',
                'commands': {
                    'serve': 'uvicore.http.commands.serve.cli',
                },
            }
        })

        # Uvicore default template options.  Because this HTTP package is high up
        # your package can easily override any of these default options!
        from uvicore.http.templating import context_functions
        self.template({
            'context_functions': {
                'url': context_functions.url,
                'asset': context_functions.asset,
                'public': context_functions.public,
            },
        })




import uvicore
from uvicore.typing import Dict
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.console.provider import Cli
from uvicore.support.module import load
from uvicore.console import group as cli_group
from uvicore.console.handlers import bootstrap
from uvicore.foundation.events import app as AppEvents


@uvicore.provider()
class Console(ServiceProvider, Cli):


    def register(self) -> None:
        # Register IoC bindings
        #from uvicore.console.console import cli
        #dump('service------------')
        # self.bind('uvicore.console.console.cli', 'uvicore.console.console.cli',
        #     aliases=['Console', 'console', 'cli', 'cli2']
        # )

        # self.bind('Console', 'uvicore.console.console.cli',
        #     aliases=['uvicore.console.console.cli', 'console', 'cli', 'cli2']
        # )

        # Register event listeners
        AppEvents.Booted.listen(bootstrap.Console)

    def boot(self) -> None:

        # Define service provider registration control
        # No - Never allow this packages registrations to be disabled from other configs

        # Define commands
        self.commands(
            group='gen',
            help='Generate New Schematics (commands, models, views...)',
            commands={
                'command': 'uvicore.console.commands.generators.command'
            }
        )



import uvicore
from starlette.background import BackgroundTask as _BackgroundTask
from starlette.templating import _TemplateResponse

# Proxy starlette response APIs
#from starlette.responses import Response as Generic
from starlette.responses import Response
from starlette.responses import PlainTextResponse as Text
from starlette.responses import HTMLResponse as HTML
from starlette.responses import JSONResponse as JSON
from starlette.responses import UJSONResponse as UJSON
from starlette.responses import RedirectResponse as Redirect
from starlette.responses import StreamingResponse as Stream
from starlette.responses import FileResponse as File

# Get our current template system from the IoC
templates = uvicore.ioc.make('uvicore.http.templating.jinja.Jinja')

@uvicore.service()
def View(
    name: str,
    context: dict = {},
    status_code: int = 200,
    headers: dict = None,
    media_type: str = None,
    background: _BackgroundTask = None,
) -> _TemplateResponse:
    return templates.TemplateResponse(
        name=name,
        context=context,
        status_code=status_code,
        headers=headers,
        media_type=media_type,
        background=background
    )



import uvicore
import inspect
from uvicore import typing
from uvicore.http.routing.api_router import ApiRouter
from uvicore.http.routing.router import Routes as Controller
from uvicore.http.request import Request
from uvicore.support import str as string
from uvicore.support.dumper import dump, dd
from uvicore.support import module
from uvicore.orm.fields import Relation
from uvicore.http.exceptions import HTTPException, PermissionDenied
#from uvicore.auth import User
from uvicore.contracts import User
from uvicore.http.routing import Guard


@uvicore.service()
class ModelRoute:
    """Dynamic Model CRUD Routes"""

    def routes(self, Model, route, path, tags, scopes):
        """Build dynamic model CRUD routes"""

        @route.get('/' + path, response_model=typing.List[Model], tags=tags, scopes=[Model.tablename + '.read'] if scopes is None else scopes)
        #async def list(include: typing.Optional[str] = '', user: User = Guard(Model.tablename + '.read')):
        async def list(request: Request, include: typing.Optional[str] = ''):
            user: User = request.user
            # The auth guard will not allow this method, but we do have to check any INCLUDES against that models permissions
            includes = include.split(',') if include else []

            self.guard_include_permissions(Model, includes, user)
            return await Model.query().include(*includes).cache().get()

        #@route.get('/' + path + '/{id}', response_model=Model, tags=tags)
        @route.get('/' + path + '/{id}', tags=tags, scopes=[Model.tablename + '.read'] if scopes is None else scopes)
        #async def get(id: typing.Any, include: typing.Optional[str] = '', user: User = Guard(Model.tablename + '.read')):
        async def get(request: Request, id: typing.Any, include: typing.Optional[str] = ''):
            user: User = request.user
            # The auth guard will not allow this method, but we do have to check any INCLUDES against that models permissions
            includes = include.split(',') if include else []

            self.guard_include_permissions(Model, includes, user)
            return await Model.query().include(*includes).cache().find(id)

    def guard_include_permissions(self, Model, includes: typing.List, user: User):
        # No includes, skip
        if not includes: return

        # User is superadmin, allow
        if user.superadmin: return

        # Loop each include and parts
        for include in includes:

            # Includes are split into dotnotation "parts".  We have to walk down each parts relations
            parts = [include]
            if '.' in include: parts = include.split('.')

            entity = Model
            for part in parts:

                # Get field for this "include part".  Remember entity here is not only Model, but a walkdown
                # based on each part (separated by dotnotation)
                field = entity.modelfields.get(part)  # Don't use modelfield() as it throws exception

                # Field not found, include string was a typo
                if not field: continue

                # Field has a column entry, which means its NOT a relation
                if field.column: continue

                # Field is not an actual relation
                if not field.relation: continue

                # Get actual relation object
                relation = field.relation.fill(field)

                # Get model permission string for this relation
                model_permissions = [relation.entity.tablename + '.read'] if scopes is None else scopes

                # User must have ALL scopes defined on the route (an AND statement)
                authorized = True
                missing_scopes = []
                for scope in model_permissions:
                    if scope not in user.permissions:
                        authorized = False
                        missing_scopes.append(scope)

                # # Check if user has permissino to child relatoin (or superadmin)
                # if model_permission not in user.permissions:
                #     # I convert model_permissins to a list for consistency with other permission denied errors
                #     # although there will always be just one model_permission in this function
                #     # raise HTTPException(
                #     #     status_code=401,
                #     #     detail="Permission denied to {}".format(str([model_permission]))
                #     # )
                #     print('be')
                #     raise PermissionDenied(model_permission)

                if not authorized:
                    raise PermissionDenied(missing_scopes)

                # Walk down each "include parts" entities
                entity = relation.entity



            # if not field: continue
            # if not field.relation: continue
            # relation: Relation = field.relation.fill(field)
            # tablename = relation.entity.tablename
            # permission = tablename + '.read'

            # # Check if user has permissino to child relatoin (or superadmin)
            # if user.superadmin == False and permission not in user.permissions:
            #     raise HTTPException(
            #         status_code=401,
            #         detail="Access denied to {}".format(tablename)
            #     )




        # @route.post('/' + path, response_model=Model, tags=tags)
        # async def post(entity: Model):
        #     return entity


        # # GET List
        # router.add_api_route(
        #     path='/' + path,
        #     methods=['GET'],
        #     tags=tags,
        #     endpoint=self.list,
        #     response_model=List[Model],
        #     name='Get all {}'.format(path),
        #     #name=path,
        #     #   name='',
        # )

        # # GET
        # router.add_api_route(
        #     path='/' + path + '/{id}',
        #     methods=['GET'],
        #     tags=tags,
        #     endpoint=self.get,
        #     response_model=Model,
        #     name='Get {} by ID'.format(path),
        #     #name=path,
        #     #name='',
        # )

        # # POST
        # router.add_api_route(
        #     path='/' + path,
        #     methods=['POST'],
        #     tags=tags,
        #     endpoint=self.post,
        #     response_model=Model,
        #     name='Create {}'.format(path),
        #     #name=path,
        #     #name='',
        # )

    # async def list(self, include: str = ''):
    #     return await Model.query().include(*include.split(',')).get()

    # async def get(self, id: Any, include: str = ''):
    #     return await Model.query().include(*include.split(',')).find(id)

    # async def post(self, entity: Model):
    #     #dump(request.__dict__)
    #     return entity
    #     #return {'name': 'hi'}
    #     #return await Model.query().include(*include.split(',')).find(id)


@uvicore.controller()
class ModelRouter(Controller):

    def __init__(self, package, scopes: typing.List = None):
        """Init override to add custom guards parameter"""
        super().__init__(package)
        self.scopes = scopes

    def register(self, route: ApiRouter):

        # Get all models in tablename sorted order from Ioc bindings
        models = self.models()

        # Get source code for ModelRouter
        modelroute = inspect.getsource(ModelRoute)
        modelroute += "ModelRoute().routes(Model, route, path, tags, scopes)"

        #from uvicore.auth.middleware import Guard
        #from uvicore.auth.models import User


        # Loop each sorted model and dynamically add a ModelRouter
        for key, binding in models.items():
            # Temp, only hashtag
            #if binding.object.tablename != 'hashtags': continue

            # Get model information from Ioc binding
            Model = binding.object
            modelname = binding.path.split('.')[-1]
            tablename = Model.tablename
            path = tablename
            tags = [string.ucbreakup(path)]
            permissions = Model.tablename

            # Dynamically instantiate ModelRouter from source code and exec
            # passing in the proper globals.  Why?  Why not just instantiate
            # the class and be done?  Becuase pydantic doesn't understand
            # type hinting that way.  For example post(entity: Model) does NOT
            # work as the type hinter errors on the Model.  I tried every other way
            # and type hinting just does not work as expected.  Even with Generics[E]
            # Not sure why exactly. This dynamic execution however works for now.
            # One caveat to this is you cannot override and EXTEND the ModelRoute
            # class.  You can override, but you must re-impliment the entire class
            # and not just extend it.
            exec(modelroute, {
                'Model': Model,
                'route': route,
                'path': path,
                'tags': tags,
                'uvicore': uvicore,
                'typing': typing,
                'User': User,
                'Guard': Guard,
                'Request': Request,
                'HTTPException': HTTPException,
                'PermissionDenied': PermissionDenied,
                'scopes': self.scopes,
            })

        # Return router
        return route

    def models(self) -> typing.OrderedDict:
        """Get all models in tablename sorted order from Ioc bindings"""
        models = typing.OrderedDict()
        unsorted_models = {}
        model_bindings = uvicore.ioc.binding(type='model', include_overrides=False)
        for key, binding in model_bindings.items():
            tablename = binding.object.tablename
            if tablename not in models:
                unsorted_models[tablename] = binding

        # Sort the OrderedDict
        #import operator
        #x = sorted(models.items(), key=lambda x: x[1])
        #x = sorted((value, key) for (key,value) in models.items())
        keys = sorted(unsorted_models.keys())
        for key in keys:
            models[key] = unsorted_models[key]
        return models
























#             list_method = """
# @router.get('/' + path, response_model=List[Model], tags=tags)
# async def list(include: str = ''):
#     return await Model.query().include(*include.split(',')).get()
# """

#             get_method = """
# @router.get('/' + path + '/{id}', response_model=Model, tags=tags)
# async def get(id: Any, include: str = ''):
#     return await Model.query().include(*include.split(',')).find(id)
# """

#             post_method = """
# @router.post('/' + path, response_model=Model, tags=tags)
# async def post(entity: Model):
#     return entity
# """

#             # Add dynamic methods
#             exec(list_method, exec_globals)
#             exec(get_method, exec_globals)
#             exec(post_method, exec_globals)



from uvicore.typing import OrderedDict

# This is the main auth config.  All items here can be overridden
# when used inside other applications.  Accessible at config('uvicore.auth')

config = {

    # --------------------------------------------------------------------------
    # Argon2 Password Hasher Configuration
    # See https://argon2-cffi.readthedocs.io/en/stable/api.html
    # --------------------------------------------------------------------------
    'hasher': {
        'type': 'ID',
        'hash_len': 16,
        'salt_len': 16,
        'encoding': 'utf-8',
        'time_cost': 2,
        'memory_cost': 102400,
        'parallelism': 8,
    },


    # --------------------------------------------------------------------------
    # Route Configuration
    # --------------------------------------------------------------------------
    # Or like so, no underscores, so in dot notation config('blog.route.prefix')
    # have to do deep merges
    'route': {
        'web_prefix': '/auth',
        'api_prefix': '/auth/api',
    },


    # --------------------------------------------------------------------------
    # Database Connections
    # --------------------------------------------------------------------------
    'database': {
        'default': 'auth',
        'connections': {
            'auth': {
                'driver': 'mysql',
                'dialect': 'pymysql',
                'host': '127.0.0.1',
                'port': 3306,
                'database': 'iam',
                'username': 'root',
                'password': 'techie',
                'prefix': None,
            },
        },
    },


    # --------------------------------------------------------------------------
    # Registration Control
    # --------------------------------------------------------------------------
    # This lets you control the service provider registrations.  If this app
    # is used as a package inside another app you might not want some things
    # registered in that context.
    # 'registers': {
    #     'web_routes': True,
    #     'api_routes': True,
    #     'middleware': True,
    #     'views': True,
    #     'assets': True,
    #     'commands': True,
    #     'models': True,
    #     'tables': True,
    #     'seeders': True,
    # },


    # --------------------------------------------------------------------------
    # Package Dependencies (Service Providers)
    #
    # Define all the packages that this package depends on.  At a minimum, only
    # the uvicore.foundation package is required.  The foundation is very
    # minimal and only depends on configuratino, logging and console itself.
    # You must add other core services built into uvicore only if your package
    # requires them.  Services like uvicore.database, uvicore.orm, uvicore.http
    # uvicore.auth...
    # --------------------------------------------------------------------------
    # 'dependencies': OrderedDict({})

}



import uvicore
from uvicore.typing import Any, Dict
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.support.module import load
from uvicore.console.provider import Cli
from uvicore.database.handlers import bootstrap
from uvicore.foundation.events import app as AppEvents
from uvicore.http.events import server as HttpEvents


@uvicore.provider()
class Database(ServiceProvider, Cli):

    def register(self) -> None:
        # Register IoC bindings
        # self.bind('Database', 'uvicore.database.db._Db',
        #     singleton=True,
        #     aliases=['database', 'db'],
        # )

        # Set uvicore.log global
        uvicore.db = uvicore.ioc.make('uvicore.database.db.Db')

        # Register event listeners
        AppEvents.Booted.listen(bootstrap.Database)
        HttpEvents.Startup.listen(uvicore.db.http_startup)

    def boot(self) -> None:

        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define db commands
        self.commands(
            group='db',
            help='Database Commands',
            commands={
                'create': 'uvicore.database.commands.db.create',
                'drop': 'uvicore.database.commands.db.drop',
                'recreate': 'uvicore.database.commands.db.recreate',
                'seed': 'uvicore.database.commands.db.seed',
                'reseed': 'uvicore.database.commands.db.reseed',
                'connections': 'uvicore.database.commands.db.connections',
            }
        )

        # Extend schematic generator commands
        self.commands(
            group='gen',
            commands={
                'table': 'uvicore.database.commands.generators.table',
                'seeder': 'uvicore.database.commands.generators.seeder',
            }
        )



from uvicore.typing import Any, Dict, Optional, List
from starlette.exceptions import HTTPException as _HTTPException
from uvicore.http import status


class HTTPException(_HTTPException):
    def __init__(self, status_code: int, detail: Any = None, headers: Optional[Dict[str, Any]] = None) -> None:
        super().__init__(status_code=status_code, detail=detail)
        self.headers = headers


class PermissionDenied(HTTPException):
    def __init__(self, permissions: Optional[List] = None, message: str = None, *, headers: Optional[Dict[str, Any]] = None):
        detail = "Permission denied"
        if permissions:
            if type(permissions) != list: permissions = [permissions]
            detail += " to {}".format(permissions)
        if message: detail += ' - ' + str(message)
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=detail,
            headers=headers,
        )


class NotAuthenticated(HTTPException):
    def __init__(self, message: str = None, *, headers: Optional[Dict[str, Any]] = None):
        detail = 'Not authenticated'
        if message: detail += ' - ' + str(message)
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=detail,
            headers=headers,
        )


class InvalidCredentials(HTTPException):
    def __init__(self, message: str = None, *, headers: Optional[Dict[str, Any]] = None):
        detail = 'Invalid credentials'
        if message: detail += ' - ' + str(message)
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=detail,
            headers=headers,
        )


class NotFound(HTTPException):
    def __init__(self, message: str = None, *, headers: Optional[Dict[str, Any]] = None):
        detail = 'Not found'
        if message: detail += ' - ' + str(message)
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=detail,
            headers=headers,
        )



from abc import ABC, abstractmethod
from typing import Dict, List, Any


class Template(ABC):

    @property
    @abstractmethod
    def env(self) -> Any: pass

    @property
    @abstractmethod
    def paths(self) -> List[str]: pass

    @property
    @abstractmethod
    def context_functions(self) -> Dict: pass

    @property
    @abstractmethod
    def context_filters(self) -> Dict: pass

    @property
    @abstractmethod
    def filters(self) -> Dict: pass

    @property
    @abstractmethod
    def tests(self) -> Dict: pass



import uvicore
from typing import Dict
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.console.provider import Cli


@uvicore.provider()
class Orm(ServiceProvider, Cli):

    def register(self) -> None:

        # ???
        # Maybe provide a self.event_strings({}) to at least define which event strings there are
        # then you can show in ./uvicore event list

        # Register events used in this package
        # self.events.register(
        #     name='uvicore.orm-{dynamic.your.model}-BeforeSave',
        #     description='Dynamic, before saving (insert or update) model to db.',
        #     dynamic=True,
        #     is_async=True
        # )
        # self.events.register(
        #     name='uvicore.orm-{your.model}-AfterSave',
        #     description='Dynamic, after saving (insert or update) model to db.',
        #     dynamic=True,
        #     is_async=True,
        # )

        # {model} is would be full app1.models.post.Post
        # uvicore.orm-app1.models.post.Post-Saving

        # Listen to all ORM saving
        #orm-*-saving

        # Listen to all post model events
        #orm-{post}-*

        def test(event, payload):
            print('hi')

        #dd(uvicore.events.event('uvicore.foundation.events.app.Registered'))

        # Register IoC bindings
        # Automatic - self.bind('Model', 'uvicore.orm.model._Model', aliases=['model'])
        # Automatic - self.bind('ModelMetaclass', 'uvicore.orm.metaclass._ModelMetaclass')

        #self.bind('Model', 'uvicore.orm.model._Model', aliases=['model'])
        #self.bind('ModelMetaclass', 'uvicore.orm.metaclass._ModelMetaclass')

    def boot(self) -> None:
        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define commands
        self.commands({
            # Extend schematic generator commands
            'gen': {
                'commands': {
                    'model': 'uvicore.orm.commands.generators.model',
                },
            }
        })



# from abc import ABC, abstractmethod

# # uvicore/
# #     foundation/
# #     database/
# #     app.py

# from uvicore import app





# ###############
# interfaces: {
#     'log': 'uvicore.log.MainLogger'
#     'log': 'mreschke.log.MyCustomLogger'

#     'db': {
#         'cli': 'uvicore.database.Database',
#         'http': 'uvicore.database.EncodeDatabase'
#     }
# }




# from uvicore import log
# log.info('hi')


# from uvicore import Database
# db = Database()


# # uvicore __init__.py
# log = import_module(interfaces['logger']).mod
# Database = import_module(interfaces['db'].mod)

# # uvicore.interfaces.__init__.py
# from .log import LogInterface


# # uvicore.interfaces.log
# class LogInterface(ABC):
#     def info(self, log):
#         pass

# from uvicore.interfaces import LogInterface
# class MainLogger(LogInterface):
#     def info(self. log):
#         print('MAIN logger info here')

# class MyCustomLogger(LogInterface):
#     def info(self, log):
#         print('MY CUSTOM logger info here')


# #contracts.py
# #from


# #from uvicore.contracts import Database
# #db = Database('constr')
# #db.select(table)
# #db.

# class DatabaseInterface(ABC):
#     """Abstract Database Interface
#     """

#     @abstractmethod
#     def example(self, something):
#         pass



#import typer_async as typer
from typing import Optional

import uvicore
from uvicore import app, db, log
from uvicore.console import argument, click, command, option
from uvicore.support import module
from uvicore.support.dumper import dd, dump

# Commands
# create = typer.Typer()
# drop = typer.Typer()
# recreate = typer.Typer()
# seed = typer.Typer()
# reseed = typer.Typer()
# connections = typer.Typer()

# Common arguments
#connection_arg = typer.Argument(..., help='Connection(s) name (comma separated for multiples)')

def get_metakeys(connections: str):
    """Convert connections into deduplicated metakeys"""
    metakeys = []
    connections = connections.split(',') if ',' in connections else [connections]
    for connection in connections:
        conn = db.connection(connection)
        if conn:
            metakey = conn.metakey
            if metakey not in metakeys:
                metakeys.append(metakey)
        else:
            exit('Connection string {} not found in configs'.format(connection))
    return metakeys


async def create_tables(connections: str):
    #log.header('Creating tables for connections: [' + connections + ']')
    metakeys = get_metakeys(connections)
    for metakey in metakeys:
        engine = db.engine(metakey=metakey)
        log.header('Creating tables in {} in topologically order'.format(metakey))
        metadata = db.metadata(metakey=metakey)
        for table in metadata.sorted_tables:
            log.item('Creating table {}'.format(str(table.name)))
        metadata.create_all(engine)
        print()

    # # Get all tables in these connections
    # tables = []
    # for metakey in metakeys:
    #     tables.append(db.tables(metakey=metakey))

    # # Default permissions per entity
    # permissions = [
    #     'create',   # Django add
    #     'read',     # Django view
    #     'update',   # Django change
    #     'delete',   # Django delete
    # ]

    # for table in tables:
    #     for permission in permissions:



async def drop_tables(connections: str):
    """Drop tables for connection(s)"""
    #log.header('Dropping tables for connections: [' + connections + ']')
    metakeys = get_metakeys(connections)
    for metakey in metakeys:
        engine = db.engine(metakey=metakey)
        log.header('Dropping tables from {} in topologically order'.format(metakey))
        metadata = db.metadata(metakey=metakey)
        for table in reversed(metadata.sorted_tables):
            log.item('Dropping table {}'.format(str(table.name)))
        metadata.drop_all(engine)
        print()


async def seed_auth_permissions():
    if 'auth' not in db.connections().keys(): return

    from uvicore.auth.models.permission import Permission
    metadata = db.metadata('auth')

    # Default permissions per entity
    permissions = [
        'create',   # Django add
        'read',     # Django view
        'update',   # Django change
        'delete',   # Django delete
    ]

    # Bulk insert each permission for each table
    bulk = []
    tables = sorted(metadata.tables.keys())  # For sorted order
    for table in tables:
        for permission in permissions:
            bulk.append(Permission(entity=table, name=table + '.' + permission))
    await Permission.insert(bulk)


async def seed_tables(connections: str):
    metakeys = get_metakeys(connections)

    # Seed auth permissions
    await seed_auth_permissions()

    # Seed tables
    ran = []
    for metakey in metakeys:
        packages = db.packages(metakey=metakey)
        for package in packages:
            for seeder in package.database.seeders:
                if seeder not in ran:
                    log.header('Seeding tables from {} in defined order'.format(seeder))
                    await module.load(seeder).object()
                    ran.append(seeder)
                    print()


@command()
@argument('connections')
async def create(connections: str):
    """Create tables for connection(s)"""
    await create_tables(connections)


@command()
@argument('connections')
async def drop(connections: str):
    """Drop tables for connection(s)"""
    await drop_tables(connections)


@command()
@argument('connections')
async def recreate(connections: str):
    """Recreate (drop/create) tables for connection(s)"""
    #log.header('Recreating (drop/create) tables for connections: [' + connections + ']')
    await drop_tables(connections)
    await create_tables(connections)


@command()
@argument('connections')
async def seed(connections: str):
    """Seed tables for connection(s)"""
    await seed_tables(connections)


@command()
@argument('connections')
async def reseed(connections: str):
    """Reseed (drop/create/seed) tables for connection(s)"""
    await drop_tables(connections)
    await create_tables(connections)
    await seed_tables(connections)


@command()
def connections():
    """Show all packages database connections"""
    log.header("All deep merged database connections from all defined packages")
    log.line()
    log.notice("Some connections share the same database which means their tables are in the same metedata space.  This is what the unique metakey denotes.")
    dump(db.connections)
























# @create.command()
# def create_cmd(connections: str = connection_arg):
#     """Create tables for connection(s)"""
#     #log.header('Creating tables for connections: [' + connections + ']')
#     metakeys = get_metakeys(connections)
#     for metakey in metakeys:
#         log.header('Creating tables in {} in topologically order'.format(metakey))
#         metadata = db.metadata(metakey=metakey)
#         for table in metadata.sorted_tables:
#             log.item('Creating table {}'.format(str(table.name)))
#         metadata.create_all(db.engine(metakey=metakey))
#         print()

# @drop.command()
# def drop_cmd(connections: str = connection_arg):
#     """Drop tables for connection(s)"""
#     #log.header('Dropping tables for connections: [' + connections + ']')
#     metakeys = get_metakeys(connections)
#     for metakey in metakeys:
#         log.header('Dropping tables from {} in topologically order'.format(metakey))
#         metadata = db.metadata(metakey=metakey)
#         for table in reversed(metadata.sorted_tables):
#             log.item('Dropping table {}'.format(str(table.name)))
#         metadata.drop_all(db.engine(metakey=metakey))
#         print()

# @recreate.command()
# def recreate_cmd(connections: str = connection_arg):
#     """Recreate (drop/create) tables for connection(s)"""
#     #log.header('Recreating (drop/create) tables for connections: [' + connections + ']')
#     drop_cmd(connections)
#     create_cmd(connections)

# @seed.command()
# def seed_cmd(connections: str = connection_arg):
#     """Seed tables for connection(s)"""
#     metakeys = get_metakeys(connections)
#     ran = []
#     for metakey in metakeys:
#         packages = db.packages(metakey=metakey)
#         for package in packages:
#             for seeder in package.seeders:
#                 if seeder not in ran:
#                     log.header('Seeding tables from {} in defined order'.format(seeder))
#                     module.load(seeder).object()
#                     ran.append(seeder)
#                     print()

# @reseed.command()
# def reseed_cmd(connections: str = connection_arg):
#     """Reseed (drop/create/seed) tables for connection(s)"""
#     recreate_cmd(connections)
#     seed_cmd(connections)

# @connections.command()
# def connections_cmd():
#     """Show all packages database connections"""
#     log.header("All deep merged database connections from all defined packages").line()
#     log.notice("Some connections share the same database which means their tables are in the same metedata space.  This is what the unique metakey denotes.")
#     dump(db.connections)



import uvicore
import sqlalchemy as sa
from uvicore.database import Table
from uvicore.support.dumper import dump, dd

# Get related tablenames with proper prefixes
users = uvicore.db.tablename('auth.users')
formats = uvicore.db.tablename('xx_appname.formats')


@uvicore.table()
class xx_TableName(Table):

    # Actual database table name
    # Plural table names and singluar model names are encouraged
    # Do not add a package prefix, leave that to the connection config
    name = 'xx_tablename'

    # Connection for this database from your config file
    connection = 'xx_appname'

    # SQLAlchemy Table definition as a list (exclude name and metadata)
    # This will be converted into an actual SQLAlchemy Table() instance
    # See https://docs.sqlalchemy.org/en/13/core/schema.html
    schema = [
        # Defaults: nullable=False, index=False, unique=False, primary_key=False

        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('slug', sa.String(length=100), unique=True),
        sa.Column('title', sa.String(length=100)),
        sa.Column('body', sa.Text()),
        sa.Column('format', sa.Integer, sa.ForeignKey(f"{formats}.id"), nullable=False),
        sa.Column('view_count', sa.Integer),
        sa.Column('is_deleted', sa.Boolean(), default=False),
        sa.Column('is_hidden', sa.Boolean(), default=False),
        sa.Column('creator_id', sa.Integer, sa.ForeignKey(f"{users}.id"), nullable=False),
        sa.Column('updator_id', sa.Integer, sa.ForeignKey(f"{users}.id"), nullable=False),
        sa.Column('created_at', sa.DateTime()),
        sa.Column('updated_at', sa.DateTime()),
        sa.Column('indexed_at', sa.DateTime()),
    ]

    # Optional SQLAlchemy Table() instance kwargs
    schema_kwargs = {}



import uvicore
from typing import Any


def url(context: dict, name: str, **path_params: Any) -> str:
    request = context["request"]
    return request.url_for(name, **path_params)

def asset(context: dict, path: str) -> str:

    # Detect custom host and path
    asset_host = uvicore.config.app.web.asset.host
    asset_path = uvicore.config.app.web.asset.path or '/assets'

    if asset_host:
        # Using a custom asset host
        return asset_host + asset_path + '/' + path
    else:
        # Use the current running servers domain as the base URL
        request = context["request"]
        return request.url_for('assets', path=path)

def public(context: dict, path: str) -> str:
    request = context["request"]
    return request.url_for('public', path=path)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['uvicore',
 'uvicore.auth',
 'uvicore.auth.authenticators',
 'uvicore.auth.commands',
 'uvicore.auth.config',
 'uvicore.auth.contractsOLD',
 'uvicore.auth.database.seeders',
 'uvicore.auth.database.tables',
 'uvicore.auth.middleware',
 'uvicore.auth.models',
 'uvicore.auth.support',
 'uvicore.auth.user_providers',
 'uvicore.cache',
 'uvicore.cache.backends',
 'uvicore.configuration',
 'uvicore.configuration.commands',
 'uvicore.console',
 'uvicore.console.commands',
 'uvicore.console.commands.stubs',
 'uvicore.console.handlers',
 'uvicore.container',
 'uvicore.container.commands',
 'uvicore.contracts',
 'uvicore.database',
 'uvicore.database.OBSOLETE',
 'uvicore.database.commands',
 'uvicore.database.commands.stubs',
 'uvicore.database.handlers',
 'uvicore.events',
 'uvicore.events.commands',
 'uvicore.factories',
 'uvicore.foundation',
 'uvicore.foundation.config',
 'uvicore.foundation.decorators',
 'uvicore.foundation.events',
 'uvicore.http',
 'uvicore.http.OBSOLETE',
 'uvicore.http.commands',
 'uvicore.http.controllers',
 'uvicore.http.events',
 'uvicore.http.middleware',
 'uvicore.http.routing',
 'uvicore.http.servers',
 'uvicore.http.templating',
 'uvicore.logging',
 'uvicore.orm',
 'uvicore.orm.commands',
 'uvicore.orm.commands.stubs',
 'uvicore.package',
 'uvicore.package.commands',
 'uvicore.redis',
 'uvicore.support',
 'uvicore.typing']

package_data = \
{'': ['*'], 'uvicore.auth': ['http/public/*', 'http/public/assets/js/*']}

install_requires = \
['aioredis>=1.3.0,<1.4.0',
 'argon2-cffi>=20.1.0,<20.2.0',
 'asyncclick>=7.1.0,<7.2.0',
 'colored>=1.4.0,<1.5.0',
 'cryptography>=3.4.0,<3.5.0',
 'environs>=9.3.0,<9.4.0',
 'prettyprinter>=0.18.0,<0.19.0']

extras_require = \
{'database': ['sqlalchemy>=1.3.0,<1.4.0',
              'alembic>=1.5.0,<1.6.0',
              'databases[sqlite,postgres,mysql]>=0.4.0,<0.5.0'],
 'web': ['fastapi>=0.63.0,<0.64.0',
         'uvicorn>=0.13.0,<0.14.0',
         'gunicorn>=20.0.0,<20.1.0',
         'aiofiles>=0.6.0,<0.7.0',
         'Jinja2>=2.11.0,<2.12.0',
         'requests>=2.25.0,<2.26.0',
         'itsdangerous>=1.1.0,<1.2.0',
         'uvloop>=0.14.0,<0.15.0',
         'httptools>=0.1.0,<0.2.0',
         'python-multipart>=0.0.0,<0.1.0',
         'PyJWT>=2.0.0,<2.1.0']}

setup_kwargs = {
    'name': 'uvicore',
    'version': '0.1.9',
    'description': 'The Async Python Framework for Artisans. An Elegant Fullstack Python Web, API and CLI Framework',
    'long_description': "# Uvicore\n\n**The Async Python Framework for Artisans. An Elegant Fullstack Python Web, API and CLI Framework**\n\n\nThis is a full stack python framework with everything included.  Built on other micro frameworks such as Starlette and FastAPI but resembling neither.  Uvicore feels like no other python framework you have ever used.  Think Laravel instead of Django or Flask.\n\nMore to come later.\n\nUnder heavy development.  Do NOT use this repository yet.\n\n\n## Features\n\n* Super fast, built on Starlette, FastAPI (but feels like neither) and other libraries.\n* Equally well suited for complete CLI apps, Rest APIs with OpenAPI Docs and Traditional web applications.\n* Fully asynchronous including CLI apps for unix socket programming and of course Websockets.\n* Automatic API generation from all ORM models without a single line of View or Controller code.\n* Full OpenAPI 3 automatically generated docs.\n* All apps are packages and all packages are apps.  A modular framework with app override capability for configs, templates, assets, routes, models and more.  Reuse your app inside other apps as libraries for a Python native micro service architecture.  Or use the automatically generated API from ORM models for a Rest based micro service architecture with no extra work.\n* Custom IoC (Inversion of Control) system to swap any concrete implementation without changing imports.\n* Asynchronous database layer for MySQL, SQLite and Postgres.\n* All new custom asynchronous ORM (NOT another django ORM clone) with support for every relationship including polymorphism.  Enjoy Laravel's Eloquent ORM?  You'll love this one.\n* Full python type hinting for IDE code intellisense across every module including ORM model fields and methods.\n\n\n\n## ToDo\n\n* Write complete ORM tests with 100% coverage\n* Need better ORM where AND with OR mixed around.  Currently it only does all ANDs then all WHEREs which won't work for complex queries.  For example grouping 2 where ANDs with an OR like so\n\n```sql\nSELECT DISTINCT posts.id, posts.unique_slug, posts.title, posts.other, posts.creator_id, posts.owner_id, attributes.*\nFROM posts LEFT OUTER JOIN attributes ON attributes.attributable_type = 'posts' AND posts.id = attributes.attributable_id\nWHERE\n(attributes.key = 'post1-test1' and attributes.value = 'value for post1-test1')\nor\n(attributes.key = 'post2-test1' and attributes.value = 'value for post2-test1')\n```\n\n* Cannot update attributes from a parent (post.add('attributes'))...IntegrityError.  See posts seeder for notes.\n",
    'author': 'Matthew Reschke',
    'author_email': 'mail@mreschke.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/uvicore/framework',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import uvicore
from uvicore.typing import Dict, List, OrderedDict, get_type_hints, Tuple
from uvicore.events import Handler
from uvicore.support import module
from uvicore.support.dumper import dump, dd
from uvicore.foundation.events.app import Booted as OnAppBooted
from uvicore.contracts import Package as Package
from uvicore.console import command_is
from starlette.applications import Starlette
from fastapi import FastAPI
from uvicore.http import response
from uvicore.http.events import server as HttpServerEvents
from uvicore.http.routing.router import Routes
from functools import partial, update_wrapper
from uvicore.http.routing import ApiRoute, WebRoute


class Http(Handler):

    def __call__(self, event: OnAppBooted):
        """Bootstrap the HTTP Server after the Application is Booted"""

        # This will never fire if the uvicore.http module is not loaded
        # This will fire even if in console mode so we can handle showing routes
        # on certain commands.

        # Import each packages router files and add their routes to the package definition
        self.build_package_routes()

        # Fire up the HTTP server only if running from HTTP
        # Notice this is below building package routes above.  This is because
        # package routes are still built with certain CLI groups like package list and http route.
        # The server however is NOT fired up if not running a full http server command
        if not uvicore.app.is_http: return

        # Merge all packages Web and Api routes
        (web_routes, api_routes) = self.merge_routes()

        # If no routes, do nothing, no HTTP server at all
        if not web_routes and not api_routes: return

        # Fire up one or multiple HTTP servers.
        # If we have both web and api routes then we will mount subservers.
        # If not, we will only use one server.
        (base_server, web_server, api_server) = self.create_http_servers(web_routes, api_routes)

        # Add global web and api specific middleware
        self.add_middleware(web_server, api_server)

        # Get configured web and api prefixes
        web_prefix = self.get_prefix('app.web.prefix')
        api_prefix = self.get_prefix('app.api.prefix')

        # Add web routes to the web server
        self.add_web_routes(web_server, web_routes, web_prefix if not base_server else '')

        # Add api routes to the api server
        self.add_api_routes(api_server, api_routes, api_prefix if not base_server else '')

        # Add web paths (public, asset, view) and configure templates
        self.configure_webserver(web_server)

        # Fire up the proper servers and set our global app.http instance
        if base_server:
            # Running both web and api servers, we must have a perent base server
            # to mount the web and api sub servers into.  This is how we properly
            # separate web vs api middleware and other configurations.
            # Longer prefix has to be FIRST to match subapp regex properly
            if len(web_prefix) > len(api_prefix):
                base_server.mount(web_prefix, web_server)  # Web first
                base_server.mount(api_prefix, api_server)
            else:
                base_server.mount(api_prefix, api_server)  # Api first
                base_server.mount(web_prefix, web_server)
            uvicore.app._http = base_server
        elif web_server:
            # We only have a web server (no api routes)
            uvicore.app._http = web_server
        elif api_server:
            # We only have an api server (no web routes)
            uvicore.app._http = api_server

        # Attach to Starlette events and translate into Uvicore events
        @uvicore.app.http.on_event("startup")
        def startup():
            #uvicore.events.dispatch('uvicore.http.events.server.Startup')
            HttpServerEvents.Startup().dispatch()

        @uvicore.app.http.on_event("shutdown")
        def shutdown():
            #uvicore.events.dispatch('uvicore.http.events.server.Shutdown')
            HttpServerEvents.Shutdown().dispatch()


        # # Experiment add custom exceptions
        # When you do, see if you can get JSON responses to api_server and HTML responses to web_server
        # And adjust the JSON from fastapi exception_handler.py to {message} not {detail} to look more like kongs return
        # async def internal_server_error_500(request, exc):
        #     print(exc)
        #     return response.HTML(content='custom 500 internal server error', status_code=500)

        # async def not_found_404(request, exc):
        #     print(exc)
        #     return response.HTML(content='custom 404 not found', status_code=505)
        # uvicore.app.http.server.add_exception_handler(500, internal_server_error_500)
        # uvicore.app.http.server.add_exception_handler(404, not_found_404)
        # #dump('xxxxxx', uvicore.app.http.server.exception_handlers)


        # Debug and dump the actual HTTP servers (base, web, api) info and routes
        debug_dump = False
        if debug_dump:
            dump('#################################################################')
            dump("Main HTTP Server APPLICATION", uvicore.app.http.__dict__)

            dump('#################################################################')
            dump("Main HTTP Server Routes")
            for route in uvicore.app.http.routes:
                dump(route.__dict__)

            for route in uvicore.app.http.routes:
                if route.name is None:
                    dump('#################################################################')
                    dump("Sub Server '{}' APPLICATION".format(route.path))
                    dump(route.app.__dict__)

                    dump('#################################################################')
                    dump("Sub Server '{}' Routes".format(route.path))
                    for r in route.app.router.routes:
                        dump(r.__dict__)

    def build_package_routes(self) -> None:
        """Import all packages web and api routes files and add to packages route definition"""
        for package in uvicore.app.packages.values():
            package.web.routes = self.import_package_web_routes(package)
            package.api.routes = self.import_package_api_routes(package)

    def import_package_web_routes(self, package: Package) -> Dict:
        """Import one package web routes and return routes Dict"""

        # If no routes_modules defined, nothing todo
        if not package.web.routes_module: return

        # Allow only if running as HTTP or from certain CLI commands like package list/show...
        if package.registers.web_routes and (
            uvicore.app.is_http or
            command_is('http') or
            command_is('package') or
            command_is('ioc')
        ) == False: return

        routes_module = package.web.routes_module
        prefix = package.web.prefix
        name_prefix = package.web.name_prefix

        # Import main web routes module
        routes: Routes = module.load(routes_module).object(package)

        # Get name prefix from package name plus custom name prefix
        if name_prefix:
            if name_prefix[0] == '.': name_prefix = name_prefix[1:]
            name_prefix = package.short_name + '.' + name_prefix
        else:
            name_prefix = package.short_name

        # Import the web router and create a new instance
        from uvicore.http.routing.web_router import WebRouter  # isort:skip
        router = WebRouter(package, prefix, name_prefix)

        # Get returned router with all defined routes
        router = routes.register(router)

        # Merge routes level middleware into each route
        routes_middleware = routes._middleware()
        if routes_middleware:
            for route in router.routes.values():
                (route.middleware, route.endpoint) = router._merge_route_middleware(routes_middleware, route.middleware, route.endpoint)

        # Return routes
        return router.routes

    def import_package_api_routes(self,  package: Package) -> Dict:
        """Import one package api routes and return routes Dict"""

        # If no routes_modules defined, nothing todo
        if not package.api.routes_module: return

        # Allow only if running as HTTP or from certain CLI commands like package list/show...
        if package.registers.api_routes and (
            uvicore.app.is_http or
            command_is('http') or
            command_is('package') or
            command_is('ioc')
        ) == False: return

        routes_module = package.api.routes_module
        prefix = package.api.prefix
        name_prefix = package.api.name_prefix

        # Import main web routes module
        routes: Routes = module.load(routes_module).object(package)

        # Get name prefix from package name plus custom name prefix
        if name_prefix:
            if name_prefix[0] == '.': name_prefix = name_prefix[1:]
            name_prefix = package.short_name + '.' + name_prefix
        else:
            name_prefix = package.short_name

        # Import the api router and create a new instance
        from uvicore.http.routing.api_router import ApiRouter  # isort:skip
        router = ApiRouter(package, prefix, name_prefix)

        # Get returned router with all defined routes
        router = routes.register(router)

        # Merge routes level middleware into each route
        routes_middleware = routes._middleware()
        if routes_middleware:
            for route in router.routes.values():
                (route.middleware, route.endpoint) = router._merge_route_middleware(routes_middleware, route.middleware, route.endpoint)

        #Return routes
        return router.routes

    def merge_routes(self) -> Tuple[Dict[str, WebRoute], Dict[str, ApiRoute]]:
        """Merge all packages web and api routes together"""
        web_routes = Dict()
        api_routes = Dict()
        for package in uvicore.app.packages.values():
            # Merging in package order ensures last package wins!
            if package.web.routes:
               web_routes.merge(package.web.routes)
            if package.api.routes:
                api_routes.merge(package.api.routes)
        return (web_routes, api_routes)

    def create_http_servers(self, web_routes: Dict[str, WebRoute], api_routes: Dict[str, ApiRoute]) -> Tuple[Starlette, FastAPI, FastAPI]:
        """Fire up one or multiple HTTP servers"""

        # If we have both web and api routes then we will submount subservers.
        # If not, we will only use one server.
        base_server: Starlette = None
        web_server: FastAPI = None
        api_server: FastAPI = None
        debug = uvicore.config('app.debug'),

        # Base server is Starlette
        if web_routes and api_routes:
            base_server = Starlette(
                debug=debug,
            )

        # Web server is FastAPI with NO OpenAPI setup
        if web_routes:
            web_server = FastAPI(
                debug=debug,
                version=uvicore.app.version,
                openapi_url=None,
                swagger_ui_oauth2_redirect_url=None,
            )

        # Api server is FastAPI with OpenAPI setup
        if api_routes:
            api_server = FastAPI(
                debug=debug,
                title=uvicore.config('app.api.openapi.title'),
                version=uvicore.app.version,
                openapi_url=uvicore.config('app.api.openapi.url'),
                swagger_ui_oauth2_redirect_url=uvicore.config('app.api.openapi.docs_url') + '/oauth2-redirect',
                docs_url=uvicore.config('app.api.openapi.docs_url'),
                redoc_url=uvicore.config('app.api.openapi.redoc_url'),
                #root_path='/api',  #fixme with trying out kong
            )
        return (base_server, web_server, api_server)

    def add_middleware(self, web_server: FastAPI, api_server: FastAPI) -> None:
        """Add global web and api middleware to their respective servers"""

        # Add global web middleware
        if web_server:
            for name, middleware in uvicore.config.app.web.middleware.items():
                cls = module.load(middleware.module).object
                web_server.add_middleware(cls, **middleware.options)

        # Add global api middleware
        if api_server:
            for name, middleware in uvicore.config.app.api.middleware.items():
                cls = module.load(middleware.module).object
                api_server.add_middleware(cls, **middleware.options)

    def add_web_routes(self, web_server, web_routes: Dict[str, WebRoute], prefix) -> None:
        """Add web routes to the web server"""
        for route in web_routes.values():
            web_server.add_api_route(
                path=(prefix + route.path) or '/',
                endpoint=route.endpoint,
                methods=route.methods,
                name=route.name,
                include_in_schema=False,
                response_class=response.HTML,
                dependencies=route.middleware,
            )
            # Starlette
            # web_server.add_route(
            #     path=route.path,
            #     route=route.endpoint,
            #     methods=route.methods,
            #     name=route.name,
            # )

    def add_api_routes(self, api_server, api_routes: Dict[str, ApiRoute], prefix) -> None:
        """Add api routes to the api server"""
        for route in api_routes.values():
            endpoint_func = route.endpoint
            response_model = route.response_model

            # If endpoint is partial, grab inside func for type hindint and docstrings
            if isinstance(route.endpoint, partial):
                # Endpoint is a partial (was overwritten to default some higher order middleware)
                # A partial overwrites the original docstring.  Functools update_wrapper will copy it back
                # as well as handle merging of other important properties
                #update_wrapper(route.endpoint, route.endpoint.func)
                endpoint_func = route.endpoint.func

                # Blank out the __doc__ on actual Partial itself, not actual endpoint inside partial.
                # If not, OpenAPI doc description will be partial(func, *args, **keywords) - new function with partial application etc...
                route.endpoint.__doc__ = None

            # Get response model from parameter or infer from endpoint return type hint
            #response_model = route.response_model if route.response_model else get_type_hints(endpoint_func).get('return')
            response_model = route.response_model or get_type_hints(endpoint_func).get('return')

            # Get openapi description from route param or endpoint docstring
            description = route.description or endpoint_func.__doc__

            # Add route
            api_server.add_api_route(
                path=(prefix + route.path) or '/',
                endpoint=route.endpoint,
                methods=route.methods,
                name=route.name,
                response_model=response_model,
                tags=route.tags,
                dependencies=route.middleware,
                summary=route.summary,
                description=description,
            )

    def configure_webserver(self, web_server) -> None:
        """Configure the webserver with public, asset and view paths and template options"""

        # Ignore if web_server was never fired up
        if not web_server: return

        # Loop each package with an HTTP definition and add to our HTTP server
        public_paths = []
        asset_paths = []
        view_paths = []
        template_options = Dict()
        for package in uvicore.app.packages.values():
            #if not 'http' in package: continue
            if not 'web' in package: continue

            # Append public paths for later
            for path in package.web.public_paths:
                path_location = module.location(path)
                if path_location not in public_paths: public_paths.append(path_location)

            # Append asset paths for later
            for path in package.web.asset_paths:
                path_location = module.location(path)
                if path_location not in asset_paths: asset_paths.append(path_location)

            # Append template paths
            for path in package.web.view_paths or []:
                view_paths.append(path)

            # Deep merge template options
            template_options.merge(package.web.template_options or {})

        # Mount all static paths
        self.mount_static(web_server, public_paths, asset_paths)

        # Initialize new template environment from the options above
        self.initialize_templates(view_paths, template_options)

    def mount_static(self, web_server, public_paths: List, asset_paths: List) -> None:
        """Mount static routes defined in all packages"""
        StaticFiles = uvicore.ioc.make('uvicore.http.static.StaticFiles')

        # Mount all asset paths
        # Last directory defined WINS, which fits our last provider wins
        asset_url = uvicore.config.app.web.asset.path or '/assets'
        web_server.mount(asset_url, StaticFiles(directories=asset_paths), name='assets')

        # Mount all public paths (always at /)
        # Since it is root / MUST be defined last or it wins above any path after it
        web_server.mount('/', StaticFiles(directories=public_paths), name='public')

    def initialize_templates(self, paths, options) -> None:
        """Initialize and configure template system"""
        # Get the template singleton from the IoC
        templates = uvicore.ioc.make('uvicore.http.templating.jinja.Jinja')

        # Add all packages view paths
        for path in paths:
            templates.include_path(module.location(path))

        # Add all packages deep_merged template options
        if 'context_functions' in options:
            for name, method in options['context_functions'].items():
                templates.include_context_function(name, method)
        if 'context_filters' in options:
            for name, method in options['context_filters'].items():
                templates.include_context_filter(name, method)
        if 'filters' in options:
            for name, method in options['filters'].items():
                templates.include_filter(name, method)
        if 'tests' in options:
            for name, method in options['tests'].items():
                templates.include_test(name, method)

        # Initialize template system
        templates.init()

    def get_prefix(self, confpath: str) -> str:
        """Get configured web and api prefixes ensuring "" prefix and no trailing /, or / if blank)"""
        prefix = uvicore.config(confpath)
        #if not prefix: prefix = '/'
        #if prefix != '/' and prefix[-1] == '/': prefix = prefix[0:-1]  # No trailing /
        if prefix and prefix[-1] == '/': prefix = prefix[0:-1]  # No trailing /
        return prefix
